// Copyright (C) 2005-2007, ENPC - INRIA - EDF R&D
// Author(s): Ir�ne Korsakissok, Yelva Roustan
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is developed in the INRIA - ENPC joint project-team CLIME and in
// the ENPC - EDF R&D joint laboratory CEREA.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, visit the Polyphemus web site:
//      http://cerea.enpc.fr/polyphemus/


#ifndef POLYPHEMUS_FILE_MODELS_BASEPOINTEMISSION_CXX


#include "BasePointEmission.hxx"


namespace Polyphemus
{


  /////////////////////
  // POINT EMISSIONS //
  /////////////////////


  //! Main constructor.
  template<class T>
  BasePointEmission<T>::BasePointEmission()
  {
  }


  //! Destructor.
  template<class T>
  BasePointEmission<T>::~BasePointEmission()
  {
    for (unsigned int i = 0; i < Emission.size(); i++)
      delete Emission[i];
  }


  //! Model initialization.
  /*! It reads the configuration and allocates memory.
   */
  template<class T>
  void BasePointEmission<T>::Init(string config_file,
				  vector<string> species_list)
  {
    ConfigStream config_stream(config_file);

    // Browses all sections "[source]" that define emissions.
    string line, type;
    int i = 0;
    while (!config_stream.IsEmpty())
      {
	line = config_stream.GetLine();
	if (split(line)[0] == "[source]")
	  {
	    config_stream.PeekValue("Type", "continuous | puff "
                                    "| temporal | continuous_line", type);
	    if (type == "continuous")
              Emission.push_back(new ContinuousEmission<T>());
	    else if (type == "puff")
              Emission.push_back(new PuffEmission<T>());
	    else if (type == "temporal")
              Emission.push_back(new TemporalEmission<T>());
	    else if (type == "continuous_line")
              Emission.push_back(new ContinuousLineEmission<T>);
	    Emission[i]->Init(config_stream, species_list);
	    i++;
	  }
      }
  }


  //! Gets the number of emissions.
  /*!
    \return The number of emissions.
  */
  template<class T>
  int BasePointEmission<T>::GetNumberEmission()
  {
    return Emission.size();
  }


  //! Gets the emission coordinates.
  /*!
    Gets the emission coordinates.
    \param abscissa the source abscissa.
    \param ordinate the source ordinate.
    \param height the source height.
    \param emission the emission index.
  */
  template<class T>
  void BasePointEmission<T>::GetEmissionCoordinates(T& abscissa, T& ordinate,
						    T& height,
						    int emission) const
  {
    Emission[emission]->GetCoordinates(abscissa, ordinate, height);
  }


  //! Gets the list of emission coordinates.
  /*!
    Gets the list of emission coordinates.
    \param coordinate_list the list of coordinates.
  */
  template<class T>
  void BasePointEmission<T>::GetEmissionCoordinates(list<Array<T, 1> >&
                                                    coordinate_list,
						    int emission)
  {
    Emission[emission]->GetCoordinates(coordinate_list);
  }


  //! Sets the emission coordinates.
  /*!
    Sets the emission coordinates.
    \param abscissa the source abscissa.
    \param ordinate the source ordinate.
    \param height the source height.
    \param emission the emission index.
  */
  template<class T>
  void BasePointEmission<T>::SetEmissionCoordinates(T abscissa, T ordinate,
						    T height, int emission)
  {
    Emission[emission]->SetCoordinates(abscissa, ordinate, height);
  }


  //! Gets the type of an emission.
  /*!
    \return The type of an emission.
  */
  template<class T>
  string BasePointEmission<T>::GetEmissionType(int emission)
  {
    return Emission[emission]->GetType();
  }


  //! Gets the beginning date of an emission.
  /*!
    \return The beginning date of an emission.
  */
  template<class T>
  Date BasePointEmission<T>::GetDateBeg(int emission)
  {
    return Emission[emission]->GetDateBeg();
  }


  //! Checks if the emission has begun at the given date.
  /*!
    \return True if the emission has begun at date date.
  */
  template<class T>
  bool BasePointEmission<T>::HasBegun(Date date, int emission)
  {
    return Emission[emission]->HasBegun(date);
  }


  //! Checks if the emission has begun at the given date.
  /*!
    \return True if the emission has begun at date date.
  */
  template<class T>
  bool BasePointEmission<T>::HasEnded(Date date, int emission)
  {
    return Emission[emission]->HasEnded(date);
  }


  //! Checks if the source is emitting during a time interval.
  /*!
    \return True if source is emitting.
  */
  template<class T>
  bool BasePointEmission<T>::IsEmitting(Date current_date,
					Date next_date,
					int emission)
  {
    return Emission[emission]->HasBegun(next_date)
      && !Emission[emission]->HasEnded(current_date);
  }


  //! Gets the list of emitted species.
  /*!
    \return The list of emitted species.
  */
  template<class T>
  vector<int> BasePointEmission<T>::GetEmittedSpeciesIndex(int emission) const
  {
    return Emission[emission]->GetEmittedSpeciesIndex();
  }


  //! Gets the species rate.
  /*!
    \return The species rate.
  */
  template<class T>
  T BasePointEmission<T>::GetRate(int emission, int species) const
  {
    return Emission[emission]->GetRate(species);
  }


  //! Gets emitted quantities for one species.
  /*!
    \return The emitted quantities for one species.
  */
  template<class T>
  void BasePointEmission<T>::GetEmission(Date date_beg, Date date_end,
					 int species, int emission,
					 Array<T, 2>& point_emission)
  {
    Emission[emission]->GetEmission(date_beg, date_end, species,
				    point_emission);
  }
  
 
  //! Checks if plume rise parameters are given for this source.
  /*!
    \return True if there is plume rise.
  */
  template<class T>
  bool BasePointEmission<T>::HasPlumeRise(int emission)
  {
    return Emission[emission]->HasPlumeRise();
  }

  
  //! Gets the source parameters needed for plume rise computation.
  /*!
    \return The source parameters needed for plume rise computation.
  */
  template<class T>
  void BasePointEmission<T>::GetPlumeRiseParam(T& velocity, T& temperature,
					       T& diameter, int emission)
  {
    Emission[emission]->GetPlumeRiseParam(velocity, temperature, diameter);
  }

  
  //! Gets the emission timespan in a time interval.
  /*!
    \return The emission timespan.
  */
  template<class T>
  T BasePointEmission<T>::GetEmissionTimespan(const Date& current_date,
					      const Date& next_date,
					      int emission) const
  {
    return Emission[emission]->GetEmissionTimespan(current_date, next_date);
  }

} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODELS_BASEPOINTEMISSION_CXX
#endif
