// Copyright (C) 2006-2007, ENPC - INRIA - EDF R&D
// Author(s): Hadjira Foudhil, Vivien Mallet, Irène Korsakissok,
// Régis Briant
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is developed in the INRIA - ENPC joint project-team CLIME and in
// the ENPC - EDF R&D joint laboratory CEREA.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, visit the Polyphemus web site:
//      http://cerea.enpc.fr/polyphemus/

// This file is part of a Gaussian plume model for Polyphemus.


#ifndef POLYPHEMUS_FILE_MODELS_GAUSSIANPLUME_CXX


#include "GaussianPlume.hxx"
#include "BriggsFormula.hxx"


namespace Polyphemus
{


  /////////////////
  // CONSTRUCTOR //
  /////////////////


  //! Main constructor.
  /*!
    \param config_file Configuration file.
  */
  template<class T>
  GaussianPlume<T>::GaussianPlume(string config_file):
    BaseModel<T>(config_file), Nsource(0), cos_angle_(0.), sin_angle_(0.)
  {
  }


  ////////////////////
  // ACCESS METHODS //
  ////////////////////


  //! Gets a Boolean option.
  /*!
    \param option the option name.
    \return The option value.
  */
  template<class T>
  bool GaussianPlume<T>::GetBooleanOption(string option)
  {
    if (option == "gillani")
      return option_gillani;
    if (option == "hpdm")
      return option_hpdm;
    if (option == "rural")
      return option_rural;
    if (option == "day")
      return option_day;
    if (option == "scavenging")
      return option_scavenging;
    if (option == "dry_deposition")
      return option_dry_deposition;
    if (option == "plume_rise")
      return option_plume_rise;
    if (option == "breakup")
      return option_breakup;
    throw string("Unknown Boolean option: \"") + option + "\".";
  }


  //! Sets a Boolean option.
  /*!
    \param option the option name.
    \param value the option value.
  */
  template<class T>
  void GaussianPlume<T>::SetBooleanOption(string option, bool value)
  {
    if (option == "gillani")
      option_gillani = value;
    else if (option == "hpdm")
      option_hpdm = value;
    else if (option == "rural")
      option_rural = value;
    else if (option == "day")
      option_day = value;
    else if (option == "scavenging")
      option_scavenging = value;
    else if (option == "dry_deposition")
      option_dry_deposition = value;
    else if (option == "plume_rise")
      option_plume_rise = value;
    else if (option == "breakup")
      option_breakup = value;
    else
      throw string("Unknown Boolean option: \"") + option + "\".";
  }


  //! Gets a string option.
  /*!
    \param option the option name.
    \return The option value.
  */
  template<class T>
  string GaussianPlume<T>::GetStringOption(string option)
  {
    if (option == "parameterization_std")
      if (option_briggs)
	return "Briggs";
      else if (option_doury)
	return "Doury";
      else
	return "similarity_theory";
    if (option == "deposition_model")
      return deposition_model;
    if (option == "stability_class")
      return stability_class_;
    throw string("Unknown string option: \"") + option + "\".";
  }


  //! Sets a string option.
  /*!
    \param option the option name.
    \param value on exit, the option value.
  */
  template<class T>
  void GaussianPlume<T>::SetStringOption(string option, string value)
  {
    if (option == "parameterization_std")
      if (value == "Briggs")
	{
	  option_briggs = true;
	  option_doury = false;
	}
      else if (value == "Doury")
	{
	  option_briggs = false;
	  option_doury = true;
	}
      else if (value == "similarity_theory")
	{
	  option_briggs = false;
	  option_doury = false;
	}
      else
	throw string("Unknown parameterization for standard deviation: \"")
	  + value + "\".";
    else if (option == "deposition_model")
      {
	deposition_model = value;
	if (deposition_model != "Chamberlain"
	    && deposition_model != "Overcamp")
	  throw string("The deposition model must be \"Chamberlain\"")
	    + " or \"Overcamp\".";
      }
    else if (option == "stability_class")
      {
	stability_class_ = value;
	if (stability_class_ == "A")
	  stability_ = 0;
	else if (stability_class_ == "B")
	  stability_ = 1;
	else if (stability_class_ == "C")
	  stability_ = 2;
	else if (stability_class_ == "D")
	  stability_ = 3;
	else if (stability_class_ == "E")
	  stability_ = 4;
	else if (stability_class_ == "F")
	  stability_ = 5;
	else
	  throw string("Unknown stability class: \"")
	    + stability_class_ + "\".";
      }
    else
      throw string("Unknown string option: \"") + option + "\".";
  }


  //! Gets a numerical value that defines the configuration of the model.
  /*!
    \param name name of the numerical value.
    \return The value.
  */
  template<class T>
  T GaussianPlume<T>::GetNumericalValue(string name)
  {
    if (name == "temperature")
      return temperature_;
    if (name == "wind_angle")
      return wind_angle_;
    if (name == "wind")
      return wind_;
    if (name == "inversion_height")
      return inversion_height_;
    if (name == "friction_velocity")
      return friction_velocity_;
    if (name == "convective_velocity")
      return convective_velocity_;
    if (name == "boundary_height")
      return boundary_height_;
    if (name == "lmo")
      return lmo_;
    if (name == "coriolis")
      return coriolis_;
    if (name == "cos_angle")
      return cos_angle_;
    if (name == "sin_angle")
      return sin_angle_;
    throw string("Unknown numerical value: \"") + name + "\".";
  }


  //! Sets a numerical value that defines the configuration of the model.
  /*!
    \param name name of the numerical value.
    \param value the new value.
  */
  template<class T>
  void GaussianPlume<T>::SetNumericalValue(string name, T value)
  {
    if (name == "temperature")
      temperature_ = value;
    else if (name == "wind_angle")
      wind_angle_ = value;
    else if (name == "wind")
      wind_ = value;
    else if (name == "inversion_height")
      inversion_height_ = value;
    else if (name == "friction_velocity")
      friction_velocity_ = value;
    else if (name == "convective_velocity")
      convective_velocity_ = value;
    else if (name == "boundary_height")
      boundary_height_ = value;
    else if (name == "lmo")
      lmo_ = value;
    else if (name == "coriolis")
      coriolis_ = value;
    else if (name == "cos_angle")
      cos_angle_ = value;
    else if (name == "sin_angle")
      sin_angle_ = value;
    else
      throw string("Unknown numerical value: \"") + name + "\".";
  }


  //! Gets source data.
  /*!
    \param name the name of the data.
    \param value on exit, the values for all sources.
  */
  template<class T>
  void GaussianPlume<T>::GetSourceData(string name, Array<T, 1>& value)
  {
    value.resize(Nsource);

    typename list<PlumeSource<T>* >::iterator iter;
    int i = 0;
    if (name == "rate")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetRate();
    else if (name == "velocity")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetVelocity();
    else if (name == "temperature")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetTemperature();
    else if (name == "z")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetZ();
    else if (name == "y")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetY();
    else if (name == "x")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetX();
    else if (name == "diameter")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	value(i) = (*iter)->GetDiameter();
    else
      throw string("Unknown source data: \"") + name + "\".";
  }


  //! Sets source data.
  /*!
    \param name the name of the data to be set.
    \param value The values.
  */
  template<class T>
  void GaussianPlume<T>::SetSourceData(string name, const Array<T, 1>& value)
  {
    typename list<PlumeSource<T>* >::iterator iter;
    int i = 0;
    if (name == "rate")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetRate(value(i));
    else if (name == "velocity")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetVelocity(value(i));
    else if (name == "temperature")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetTemperature(value(i));
    else if (name == "z")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetZ(value(i));
    else if (name == "y")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetY(value(i));
    else if (name == "x")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetX(value(i));
    else if (name == "diameter")
      for (iter = SourceList.begin(); iter != SourceList.end(); iter++, i++)
	(*iter)->SetDiameter(value(i));
    else
      throw string("Unknown source data: \"") + name + "\".";
  }


  //! Gets species data.
  /*!
    \param name the name of the data.
    \param value on exit, the values.
  */
  template<class T>
  void GaussianPlume<T>::GetSpeciesData(string name, Array<T, 1>& value)
  {
    if (name == "half_life_time")
      value = half_life_time;
    else if (name == "biological_half_life_time")
      value = biological_half_life_time;
    else if (name == "scavenging_coefficient")
      value = scavenging_coefficient;
    else if (name == "deposition_velocity")
      value = deposition_velocity;
    else
      throw string("Unknown species data: \"") + name + "\".";
  }


  //! Sets species data.
  /*!
    \param name the name of the data to be set.
    \param value The values.
  */
  template<class T>
  void GaussianPlume<T>::SetSpeciesData(string name, const Array<T, 1>& value)
  {
    if (name == "half_life_time")
      half_life_time = value;
    else if (name == "biological_half_life_time")
      biological_half_life_time = value;
    else if (name == "scavenging_coefficient")
      scavenging_coefficient = value;
    else if (name == "deposition_velocity")
      deposition_velocity = value;
    else
      throw string("Unknown species data: \"") + name + "\".";
  }


  /////////////////////
  // INITIALIZATIONS //
  /////////////////////


  //! Reads the configuration.
  /*! It reads the description of the domain, the simulation starting-date,
    species lists, options (especially which processes are included) and the
    paths to data input-files.
  */
  template<class T>
  void GaussianPlume<T>::ReadConfiguration()
  {
    BaseModel<T>::ReadConfiguration();

    this->config.SetSection("[domain]");

    // Land category ("rural" or "urban").
    string land_category;
    this->config.PeekValue("Land_category", "rural | urban", land_category);
    option_rural = land_category == "rural";

    // Time (Night or Day).
    string day_night;
    this->config.PeekValue("Time", "day | night", day_night);
    option_day = day_night == "day";

    /*** Gaussian options ***/

    this->config.SetSection("[gaussian]");

    this->config.PeekValue("With_radioactive_decay",
			   option_radioactive_decay);
    this->config.PeekValue("With_biological_decay",
			   option_biological_decay);
    this->config.PeekValue("With_scavenging",
			   option_scavenging);
    this->config.PeekValue("With_dry_deposition",
			   option_dry_deposition);
    this->config.PeekValue("With_plume_rise",
			   option_plume_rise);
    this->config.PeekValue("File_source", file_source);
    this->config.PeekValue("File_correction", correction_coefficient_path_);

    // Parameterization to compute standard deviations ("Briggs", "Doury").
    string sigma_formula, sigma_formula_above;
    this->config.PeekValue("Sigma_parameterization",
			   "Briggs | Doury | similarity_theory",
			   sigma_formula);
    option_briggs = sigma_formula == "Briggs";
    option_doury = sigma_formula == "Doury";

    // Parameterization for standard deviation above BL.
    this->config.PeekValue("Above_BL", "Gillani | none", sigma_formula_above);
    option_gillani = sigma_formula_above == "Gillani";

    // Alternative formulae for similarity theory (for elevated sources).
    if (sigma_formula == "similarity_theory" || read_all)
      this->config.PeekValue("With_HPDM", option_hpdm);

    // Plume rise parameterization.
    if (option_plume_rise || read_all)
      {
	this->config.PeekValue("With_plume_rise_breakup", option_breakup);
	string plume_rise_formula;
	this->config.PeekValue("Plume_rise_parameterization",
			       "HPDM | Holland | Concawe",
			       plume_rise_formula);
	option_holland = plume_rise_formula == "Holland";
	option_concawe = plume_rise_formula == "Concawe";
      }

    // Gets the deposition model.
    if (option_dry_deposition || read_all)
      {
	this->config.SetSection("[deposition]");
	this->config.PeekValue("Deposition_model", "Chamberlain | Overcamp",
			       deposition_model);

	if (deposition_model == "Chamberlain")
	  this->config.PeekValue("Nchamberlain", "> 0", Nchamberlain);
      }
    else
      Nchamberlain = 0;

    /*** Species data ***/
    ConfigStream species_stream(this->file_species);

    // Reads the species half-lives in the species file.
    if (option_radioactive_decay)
      {
	half_life_time.resize(this->Ns);

	species_stream.SetSection("[half_life]");
	for (int i = 0; i < this->Ns; i++)
	  species_stream.PeekValue(this->species_list[i], half_life_time(i));
	
	// Conversion from days to seconds.
	T factor = 24. * 3600.;
	for (int i = 0; i < this->Ns; i++)
	  half_life_time(i) *= factor;
      }

    // Reads the species biological half-lives in the species file.
    if (option_biological_decay)
      {
	biological_half_life_time.resize(this->Ns);

	T day_value, night_value;
	for (int i = 0; i < this->Ns; i++)
	  {
	    species_stream.SetSection("[half_life_time]");
	    species_stream.Find(this->species_list[i]);
	    species_stream.GetNumber(day_value);
	    species_stream.GetNumber(night_value);
	    if (option_day)
	      biological_half_life_time(i) = day_value;
	    else
	      biological_half_life_time(i) = night_value;
	  }
      }
  }


  //! Allocates memory.
  /*! Allocates the grids and the concentration Data for gaseous species.
   */
  template<class T>
  void GaussianPlume<T>::Allocate()
  {
    BaseModel<T>::Allocate();
  }


  //! Sources initialization.
  /*! It sets all sources from a text file. Each source is described in a
    dedicated section "[source]" in which one finds the following entries:
    <ul>
    <li> Rate: the rate (mass unit per second),
    <li> Velocity: the efflux speed (m/s),
    <li> Temperature: the temperature of emissions (Celsius degrees),
    <li> Abscissa: abscissa (m),
    <li> Ordinate: ordinate (m),
    <li> Altitude: height (m),
    <li> Species_name: species name.
    </ul>
    \param source_file file that describes the sources.
  */
  template<class T>
  void GaussianPlume<T>::InitSource(string source_file)
  {
    // List of emissions.
    this->PointEmissionManager = new BasePointEmission<T>();
    this->PointEmissionManager->Init(source_file, this->species_list);

    T rate, velocity, temperature, diameter, x1, y1, z1;
    string species_name;
    SourceList.clear();
    Nsource = 0;

    int Nemis = this->PointEmissionManager->GetNumberEmission();
    for (int emission = 0; emission < Nemis; emission++)
      {
	vector<int> emitted_species_index =
	  this->PointEmissionManager->GetEmittedSpeciesIndex(emission);
	int Ns_emitted = emitted_species_index.size();
	int s;

	// Loop on all emitted species.
	for (int species = 0; species < Ns_emitted; species++)
	  {
	    s = emitted_species_index[species];
	    rate = this->PointEmissionManager->GetRate(emission, species);
	    if (this->PointEmissionManager->GetEmissionType(emission)
                == "continuous")
	      {
		this->PointEmissionManager->
		  GetPlumeRiseParam(velocity, temperature, diameter,
                                    emission);
 		this->PointEmissionManager
 		  ->GetEmissionCoordinates(x1, y1, z1, emission);
		PlumeSource<T>* ContinuousSource =
                  new PlumeSource<T>(rate, velocity, temperature, diameter,
                                     x1, y1, z1, s);
 		SourceList.push_back(ContinuousSource);
                Nsource++;
  	      }
	    else if (this->PointEmissionManager->GetEmissionType(emission)
                     == "continuous_line")
	      {
		this->PointEmissionManager->
		  GetPlumeRiseParam(velocity, temperature, diameter,
                                    emission);
		list<Array<T, 1> > coordinate_list;
		Array<T, 1> coordinate(6);
		this->PointEmissionManager
		  ->GetEmissionCoordinates(coordinate_list, emission);
		for (typename list<Array<T, 1> >::const_iterator iter
		       = coordinate_list.begin();
                     iter != coordinate_list.end(); iter++)
		  {
	      	    coordinate = *iter;
		    PlumeSource<T>* LineSource =
                      new PlumeLineSource<T>(rate, coordinate(0),
                                             coordinate(1), coordinate(2),
                                             coordinate(3), coordinate(4),
                                             coordinate(5), diameter, s);
		    SourceList.push_back(LineSource);
                    Nsource++;
		  }
	      }
	    else
              throw string("Source type must be \"continuous\" or ")
                + "\"continuous_line\" but is "
                + this->PointEmissionManager->GetEmissionType(emission);
	  }
      }
    Nsource_temp = Nsource;
  }


  //! Model initialization.
  /*! It reads the configuration.
    \param read_all_input Should all input data and options be read in the
    configuration files? If so, even if unnecessary data will be read,
    e.g. the Monin-Obukhov length while the Doury parameterization is
    used. This enables to switch et a later time to other parameterizations
    (e.g., to "similarity_theory").
  */
  template<class T>
  void GaussianPlume<T>::Init(bool read_all_input = false)
  {
    read_all = read_all_input;

    BaseModel<T>::Init();

    /*** Initializes the sources ***/

    InitSource(file_source);
  }


  //! Initializes meteorological conditions.
  /*! It sets the meteorological data from a configuration file.  The
    situation is described in a dedicated section "[situation]" in which one
    finds the following entries:
    <ul>
    <li> Temperature: the temperature of ambient air (Celsius degrees).
    <li> Wind_angle: the angle between positive x-axis and wind,
    counterclockwise (degrees).
    <li> Wind: the wind velocity (m/s).
    <li> Inversion_height: the inversion height (m).
    <li> Stability: stability class in [A, F].
    </ul>
    \param meteo ConfigStream instance through which all entries (Temperature,
    Wind_angle, ...) may be read to set the meteorological situation.
    \param show_meteo indicates whether the meteorological data is to be
    displayed on screen.
  */
  template<class T>
  void GaussianPlume<T>::InitMeteo(ConfigStream& meteo, bool show_meteo)
  {
    meteo.PeekValue("Temperature", temperature_);
    meteo.PeekValue("Wind_angle", wind_angle_);
    meteo.PeekValue("Wind", wind_);
    meteo.PeekValue("Boundary_height", "positive", boundary_height_);
    if (option_briggs || option_doury || read_all)
      {
	if (option_plume_rise && !read_all)
	  if (option_breakup && !option_holland)
	    {
	      meteo.PeekValue("Friction_velocity", friction_velocity_);
	      meteo.PeekValue("Convective_velocity", convective_velocity_);
	    }
	meteo.PeekValue("Stability", "A | B | C | D | E | F",
			stability_class_);
	stability_class_ = upper_case(stability_class_);
	if (stability_class_ == "A")
	  stability_ = 0;
	else if (stability_class_ == "B")
	  stability_ = 1;
	else if (stability_class_ == "C")
	  stability_ = 2;
	else if (stability_class_ == "D")
	  stability_ = 3;
	else if (stability_class_ == "E")
	  stability_ = 4;
	else if (stability_class_ == "F")
	  stability_ = 5;
      }
    if (!option_briggs && !option_doury || read_all)
      {
	meteo.PeekValue("Friction_velocity", friction_velocity_);
	meteo.PeekValue("Convective_velocity", convective_velocity_);
	meteo.PeekValue("LMO", lmo_);
	meteo.PeekValue("Coriolis", coriolis_);
      }

    if (show_meteo)
      cout << "\t" << temperature_
	   << "\t\t" << wind_angle_
	   << "\t\t" << wind_
	   << "\t\t" << stability_class_ << endl;

    temperature_ = temperature_ + 273.15;

    /*** Scavenging coefficients ***/

    if (option_scavenging)
      InitScavenging(meteo);

    /*** Deposition velocities ***/

    if (option_dry_deposition)
      InitDeposition(meteo);
  }


  /*! Initializes scavenging coefficients.
    \param meteo ConfigStream instance through which scavenging coefficients
    may be read.
  */
  template<class T>
  void GaussianPlume<T>::InitScavenging(ConfigStream& meteo)
  {

    scavenging_coefficient.resize(this->Ns);
    meteo.Find("Scavenging_coefficient");
    vector<string> scav_coefficient = split(meteo.GetLine());
    vector<string>::iterator iter;
    for (int i = 0; i < this->Ns; i++)
      {
	iter = find(scav_coefficient.begin(), scav_coefficient.end(),
		    this->species_list[i]);
	if (iter++ == scav_coefficient.end() ||
	    iter == scav_coefficient.end())
	  throw string("Unable to find scavenging coefficient for")
	    + string(" species \"") + this->species_list[i] + "\".";
	scavenging_coefficient(i) = to_num<T>(*iter);
      }
  }


  /*! Initializes deposition velocities.
    \param meteo ConfigStream instance through which deposition velocities
    may be read.
  */
  template<class T>
  void GaussianPlume<T>::InitDeposition(ConfigStream& meteo)
  {
    deposition_velocity.resize(this->Ns);
    meteo.Find("Deposition_velocity");
    vector<string> dep_velocity = split(meteo.GetLine());
    vector<string>::iterator iter;
    for (int  i = 0; i < this->Ns; i++)
      {
	iter = find(dep_velocity.begin(), dep_velocity.end(),
		    this->species_list[i]);
	if (iter++ == dep_velocity.end() ||
	    iter == dep_velocity.end())
	  throw string("Unable to find deposition velocity for")
	    + string(" species \"") + this->species_list[i] + "\".";
	deposition_velocity(i) = to_num<T>(*iter);
      }
  }


  //////////////////
  // COMPUTATIONS //
  //////////////////


  //! Computations performed before any concentration can be computed.
  /*!  After this method is called, one may call 'GetConcentration' or
    'Compute'. But never before!
  */
  template<class T>
  void GaussianPlume<T>::InitCompute()
  {
    const T pi = 3.14159265358979323846264;
    T wind_angle_rad = wind_angle_ / 180. * pi;
    cos_angle_ = cos(wind_angle_rad);
    sin_angle_ = sin(wind_angle_rad);

    // If some sources were added during the previous meteo situation.
    if(Nsource_temp != Nsource)
      {
        this->EraseSource();
        this->RestoreRate();
      }

    if (!option_briggs && !option_doury)
      // Matching between LMO and stability classes.
      ComputeStabilityClass(lmo_, stability_, stability_class_);
    if (stability_ != 4 && stability_ != 5)
      inversion_height_ = boundary_height_;
    else
      inversion_height_ = 0.;

    // If the wind is parallel to the road (plus or minus 10 degrees).
    if(abs(sin_angle_) >= sin(80. * pi / 180))
      this->Discretization();

    if (option_plume_rise)
      ComputePlumeRise();

    // Correction coefficients are read only if there is a line source.
    bool coefficients_read = 0;
    for (typename list<PlumeSource<T>* >::const_iterator iter
	   = SourceList.begin();
	 iter != SourceList.end(); iter++)
      if (!coefficients_read)
        if((*iter)->GetEmissionType() == "continuous_line")
          {
            InitCorrectionCoefficients();
            coefficients_read = 1;
          }
  }

 
  //! Computes plume rise.
  template<class T>
  void GaussianPlume<T>::ComputePlumeRise()
  {
    // Loop over all sources.
    for (typename list<PlumeSource<T>* >::iterator iter
	   = SourceList.begin();
	 iter != SourceList.end(); iter++)
      if ((*iter)->GetEmissionType() == "continuous")
        ComputeSourcePlumeRise(**iter);
  }

 
  //! Computes plume rise for a given source.
  template<class T>
  void GaussianPlume<T>::ComputeSourcePlumeRise(PlumeSource<T>& source)
  {
    // Computing the effective source height.
    T source_height = source.GetZ();
    T plume_rise;
    if (option_holland)
      plume_rise = ComputeHollandPlumeRise(temperature_, wind_,
					   source.GetVelocity(),
					   source.GetTemperature(),
					   source.GetDiameter());
    else if (option_concawe)
      plume_rise = ComputeConcawePlumeRise(temperature_, wind_,
					   inversion_height_,
					   convective_velocity_,
					   friction_velocity_,
					   stability_,
					   source.GetVelocity(),
					   source.GetTemperature(),
					   source.GetDiameter(),
					   source_height, option_breakup);
    else
      plume_rise = ComputeHPDMPlumeRise(temperature_, wind_,
					inversion_height_,
					convective_velocity_,
					friction_velocity_,
					stability_,
					source.GetVelocity(),
					source.GetTemperature(),
					source.GetDiameter(),
					source_height, option_breakup);

    // Computing the additional plume spread due to plume rise.
    T initial_sigma_y_2 = source.GetSigma_y();
    T initial_sigma_z_2 = source.GetSigma_z();
    T sigma_y_pr = plume_rise / 3.5;
    T sigma_z_pr;
    if (option_gillani)
      {
	T dTdz;
	ComputePotentialTemperatureGradient(stability_, dTdz);
	dTdz = dTdz - 0.986;
	sigma_z_pr = 15. * exp(- 117. * dTdz);
      }
    else
      sigma_z_pr = plume_rise / 2.;
    source.SetSigma_y(initial_sigma_y_2 + sigma_y_pr * sigma_y_pr);
    source.SetSigma_z(initial_sigma_z_2 + sigma_z_pr * sigma_z_pr);

    // Computing effective height.
    T effective_height = 0.;
    T effective_height_above = 0.;
    T penetration_factor = 0.;
    
    if (inversion_height_ <= 0.)
      {
	// There is no inversion.
	effective_height = source_height + plume_rise;
	penetration_factor = 0.;
      }
    else if (source_height + 0.5 * plume_rise >= inversion_height_)
      {
	// Plume is above inversion layer.
	penetration_factor = 1.;
	effective_height = 0.;
	effective_height_above = source_height + plume_rise;
      }
    else if (source_height + 1.5 * plume_rise >= inversion_height_)
      {
	// Partial penetration in the inversion layer.
	penetration_factor
	  = 1.5 - (inversion_height_ - source_height) / plume_rise;
	plume_rise = (0.62 + 0.38 * penetration_factor) *
	  (inversion_height_ - source_height);
	effective_height = source_height + plume_rise;
	effective_height_above = source_height +
	  (1.38 + 0.38 * penetration_factor)
	  * (inversion_height_ - source_height);
      }
    else
      {
	// Plume is inside the boundary layer.
	effective_height = source_height + plume_rise;
	penetration_factor = 0.;
      }
    source.SetPenetrationFactor(penetration_factor);
    source.SetHeight(effective_height);
    source.SetHeightAboveBL(effective_height_above);
  }
  

  //! Computes concentration of a given species at a given point.
  /*!
    \param species species index.
    \param x abscissa (m).
    \param y ordinate (m).
    \param z height (m).
    \return The value of concentration at the given point (mass/m^3).
  */
  template<class T>
  T GaussianPlume<T>::GetConcentration(int species, T z, T y, T x)
  {
    const T pi = 3.14159265358979323846264;

    // Distances downwind and crosswind from the source (m).
    T distance_x, distance_x1, distance_x2, distance_y, distance_y1,
      distance_y2, time;
    distance_x = 0;

    // Horizontal and vertical standard deviations (m).
    T sigma_y, sigma_z, initial_sigma_y_2, initial_sigma_z_2;

    // Effective height of release (m), inside and above BL.
    T effective_height, effective_height_above;

    // Fraction of the plume that penetrates above BL.
    T penetration_factor;

    // Rate (g/s).
    T rate;

    // Output concentration.
    T concentration = 0.;

    // Loss factors.
    T loss_factor, overcamp_factor;

    // Loop over all sources.
    for (typename list<PlumeSource<T>* >::const_iterator iter
	   = SourceList.begin();
	 iter != SourceList.end(); iter++)
      if ((*iter)->GetSpeciesIndex() == species)
        {
          // Minimum volume.
          T minimum_volume = pi * (*iter)->GetDiameter()
            * (*iter)->GetDiameter() * (*iter)->GetVelocity();

          if((*iter)->GetEmissionType() == "continuous")
            {
              // Downwind distance from source.
              distance_x = (x - (*iter)->GetX()) * cos_angle_
                + (y - (*iter)->GetY()) * sin_angle_;

              // Distance from downwind axis.
              distance_y = ((*iter)->GetX() - x) * sin_angle_
                + (y - (*iter)->GetY()) * cos_angle_;
              distance_y = abs(distance_y);

              // Effective height.
              effective_height = (*iter)->GetHeight();
              penetration_factor = (*iter)->GetPenetrationFactor();
              effective_height_above = (*iter)->GetHeightAboveBL();

              // If there is anything to compute.
              if (distance_x > 0)
                {
                  // Initial plume spread.
                  initial_sigma_y_2 = (*iter)->GetSigma_y();
                  initial_sigma_z_2 = (*iter)->GetSigma_z();

                  if (effective_height != 0.)
                    {
                      // Computing standard deviations.
                      time = distance_x / wind_;
                      ComputeSigma(distance_x, time, effective_height,
                                   sigma_y, sigma_z);
 
                      sigma_y = sqrt(sigma_y * sigma_y + initial_sigma_y_2);
                      if (option_gillani && effective_height
                          >= boundary_height_)
                        {
                          initial_sigma_z_2 = max(initial_sigma_z_2, 9.);
                          sigma_z = sqrt(initial_sigma_z_2
                                         * (1. + 2.3 * sqrt(time)));
                        }
                      else
                        sigma_z = sqrt(sigma_z * sigma_z
                                       + initial_sigma_z_2);

                      // Computing loss factor.
                      ComputeLossFactor(distance_x, effective_height,
                                        species, loss_factor,
                                        overcamp_factor);

                      rate = (*iter)->GetRate() * (1. - penetration_factor);
                      concentration += loss_factor
                        * ComputePlumeConcentration(wind_, inversion_height_
                                                    , effective_height, rate
                                                    , distance_y, z,
                                                    sigma_y, sigma_z,
                                                    overcamp_factor,
                                                    minimum_volume);

                    }
                  // Fraction of the plume above the boundary layer.
                  if (effective_height_above != 0.)
                    {
                      // Computing standard deviations.
                      time = distance_x / wind_;
                      ComputeSigma(distance_x, time, boundary_height_,
                                   sigma_y, sigma_z);
                      sigma_y = sqrt(sigma_y * sigma_y + initial_sigma_y_2);
                      if (option_gillani)
                        {
                          initial_sigma_z_2 = max(initial_sigma_z_2, 9.);
                          sigma_z = sqrt(initial_sigma_z_2
                                         * (1. + 2.3 * sqrt(time)));
                        }
                      else
                        sigma_z = sqrt(sigma_z * sigma_z +
                                       initial_sigma_z_2);
			
                      // Computing loss factor.
                      ComputeLossFactor(distance_x, effective_height_above,
                                        species, loss_factor,
                                        overcamp_factor);
			
                      rate = (*iter)->GetRate() * penetration_factor;
                      concentration += loss_factor
                        * ComputePlumeConcentration(wind_, inversion_height_
                                                    , effective_height_above
                                                    , rate, distance_y, z,
                                                    sigma_y, sigma_z,
                                                    overcamp_factor,
                                                    minimum_volume);
                    }
                }
            }
          else if ((*iter)->GetEmissionType() == "continuous_line")
            {
              // Coordinates of the line source in the initial system of
              // coordinate.
              T x1, x2, y1, y2, z1, z2;

              // Coordinates in the source system of coordinate.
              T x_source, y_source, y1_source, y2_source, x1_source,
                x2_source;

              // Cosine and sine of the angle between initial axis and
              // source axis.
              T cos_gamma, sin_gamma;

              // Horizontal and vertical standard deviations (m).
              T sigma_y, sigma_z, sigma_y1 = 0, sigma_y2 = 0, sigma_z1 = 0,
                sigma_z2 = 0;

              x1 = (*iter)->GetX();
              x2 = (*iter)->GetX2();
              y1 = (*iter)->GetY();
              y2 = (*iter)->GetY2();
              z1 = (*iter)->GetZ();
              z2 = (*iter)->GetZ2();
		
              // First rotation of axis, to be in the source system of
              // coordinate.

              // The line source length must not be 0.
              T length = sqrt((x2 - x1) * (x2 - x1) +
                              (y2 - y1) * (y2 - y1) +
                              (z2 - z1) * (z2 - z1));

              // If both extremities of the source are at the same height.
              if (z1 == z2)
                {
                  if(length != 0)
                    {
                      cos_gamma = abs(y2 - y1) / length;
                      sin_gamma = abs(x2 - x1) / length;
                    }
                  else
                    throw string("There is a line source that has its")
                      + " length equal to 0.";

                  // If theta is between 0 and -90 degrees.
                  if(x2 >= x1 && y2 > y1)
                    sin_gamma = -sin_gamma;

                  // Else if theta is between -90 and -180 degrees.
                  else if(x2 >= x1 && y2 <= y1)
                    {
                      cos_gamma = -cos_gamma;
                      sin_gamma = -sin_gamma;
                    }
                  // Else if theta is between 90 and 180 degrees.
                  else if(x2 < x1 && y2 <= y1)
                    cos_gamma = -cos_gamma;

                  // Rotation and translation of axis into the source system
                  // of coordinates.
                  x1_source = 0;
                  y1_source = 0;
                  x2_source = (x2 - x1) * cos_gamma + (y2 - y1) * sin_gamma;
                  y2_source = (x1 - x2) * sin_gamma + (y2 - y1) * cos_gamma;
                  x_source = (x - x1) * cos_gamma + (y - y1) * sin_gamma;
                  y_source = (x1 - x) * sin_gamma + (y - y1) * cos_gamma;
                }
              else
                throw string("The line source must have the same height")
                  + " at both extremities";

              // theta = angle_ - gamma: wind angle in the source system
              // of coordinates.
              T cos_theta, sin_theta, tan_theta = 0;
              cos_theta = cos_angle_ * cos_gamma + sin_angle_ * sin_gamma;
              sin_theta = sin_angle_ * cos_gamma - cos_angle_ * sin_gamma;
              if(cos_theta != 0)
                tan_theta = sin_theta / cos_theta;

              // If theta is equal to -90 or 90 degree.
              if(abs(cos_theta) < 0.00000001)
                {
                  // then theta = -89 degree or 89 degree.
                  cos_theta = 0.017452406437283376;
                  if (sin_theta > 0)
                    sin_theta = 0.99984769515639127;
                  else
                    sin_theta = -0.99984769515639127;
                }

              // Distance from the point to the first extremity of the
              // source.
              distance_x1 = (x_source - x1_source) * cos_theta +
                (y_source - y1_source) * sin_theta;
              distance_y1 = (x1_source - x_source) * sin_theta
                + (y_source - y1_source) * cos_theta;
              distance_y1 = abs(distance_y1);

              // Distance from the point to the second extremity of the
              // source.
              distance_x2 = (x_source - x2_source) * cos_theta +
                (y_source - y2_source) * sin_theta;
              distance_y2 = (x2_source - x_source) * sin_theta
                + (y_source - y2_source) * cos_theta;
              distance_y2 = abs(distance_y2);

              // Downwind distance from source.
              distance_x = x_source / cos_theta;

              // Effective height.
              effective_height = (*iter)->GetHeight();
              penetration_factor = (*iter)->GetPenetrationFactor();
              effective_height_above = (*iter)->GetHeightAboveBL();

              // If there is anything to compute.
              if (distance_x > 0)
                {
                  // Initial plume spread.
                  initial_sigma_y_2 = (*iter)->GetSigma_y();
                  initial_sigma_z_2 = (*iter)->GetSigma_z();

                  if (effective_height != 0.)
                    {
                      if(distance_x1 >= 0)
                        {
                          // Computing standard deviation sigma_y1, with
                          // distance_x1.
                          time = distance_x1 / wind_;
                          ComputeSigma(distance_x1, time, effective_height,
                                       sigma_y1, sigma_z1);
                          sigma_y1 = sqrt(sigma_y1 * sigma_y1 +
                                          initial_sigma_y_2);
                        }

                      if(distance_x2 >= 0)
                        {
                          // Computing standard deviation sigma_y2, with
                          // distance_x2.
                          time = distance_x2 / wind_;
                          ComputeSigma(distance_x2, time, effective_height,
                                       sigma_y2, sigma_z2);
                          sigma_y2 = sqrt(sigma_y2 * sigma_y2 +
                                          initial_sigma_y_2);
                        }

                      // Computing standard deviations sigma_y and sigma_z,
                      // with distance_x.
                      time = distance_x / wind_;
                      ComputeSigma(distance_x, time, effective_height,
                                   sigma_y, sigma_z);

                      if (option_gillani && effective_height
                          >= boundary_height_)
                        {
                          initial_sigma_z_2 = max(initial_sigma_z_2, 9.);
                          sigma_z = sqrt(initial_sigma_z_2
                                         * (1. + 2.3 * sqrt(time)));
                        }
                      else
                        sigma_z = sqrt(sigma_z * sigma_z +
                                       initial_sigma_z_2);

                      // Computing loss factor.
                      ComputeLossFactor(distance_x, effective_height,
                                        species, loss_factor,
                                        overcamp_factor);

                      rate = (*iter)->GetRate() * (1. - penetration_factor);
                      concentration += loss_factor
                        * ComputePlumeLineConcentration(wind_,
                                                        inversion_height_,
                                                        effective_height,
                                                        rate, x_source,
                                                        y_source, z,
                                                        sigma_y1, sigma_y2,
                                                        sigma_z,
                                                        overcamp_factor,
                                                        y1_source,
                                                        y2_source,
                                                        cos_theta,
                                                        sin_theta,
                                                        minimum_volume);
                    }
                }

              if (option_briggs)
                {
                  // Downwind correction.
                  concentration = concentration /
                    (ComputeCorrection(concentration, cos_theta, sin_theta,
                                       tan_theta, y1_source, y2_source,
                                       y_source, x_source, distance_x,
                                       (*iter)->GetHeight(),
                                       (*iter)->GetSigma_y()) + 1);

                  // Extremities correction.
                  if (abs(cos_theta) <=
                      cos(threshold_extremities_correction_ * pi / 180.))
                    {
                      T extremity_correction_1 = 0,
                        extremity_correction_2 = 0;
                      extremity_correction_1 =
                        ComputeExtremitiesCorrection(rate,loss_factor, wind_,
                                                     inversion_height_,
                                                     effective_height,
                                                     distance_x1, distance_y1,
                                                     z, (*iter)->GetSigma_z(),
                                                     overcamp_factor,
                                                     minimum_volume,
                                                     abs(cos_theta), species);
                      if (concentration - extremity_correction_1 < 0)
                        extremity_correction_1 = concentration;

                      extremity_correction_2 =
                        ComputeExtremitiesCorrection(rate,loss_factor, wind_,
                                                     inversion_height_,
                                                     effective_height,
                                                     distance_x2, distance_y2,
                                                     z,(*iter)->GetSigma_z(),
                                                     overcamp_factor,
                                                     minimum_volume,
                                                     abs(cos_theta),
                                                     species);
                      if (concentration - extremity_correction_2 < 0)
                        extremity_correction_2 = concentration;
                      
                      concentration = concentration + extremity_correction_2
                        - extremity_correction_1;
                    }
                }
            }
        }
    return concentration;
  }

  //!  Computes the horizontal and vertical standard deviations.
  template<class T>
  void GaussianPlume<T>::ComputeSigma(T distance, T transfer_time,
				      T z, T& sigma_y,
				      T& sigma_z, bool diff)
  {
    // Horizontal diffusion parameter.
    if (option_briggs)
      {
	if (option_rural)
	  sigma_y =
	    ComputeRuralPlumeHorizontalSigma(distance, stability_);
	else
	  sigma_y =
	    ComputeUrbanPlumeHorizontalSigma(distance, stability_);
      }
    else if (option_doury)
      {
	if (option_day)
	  sigma_y =
	    ComputeDouryNormalPlumeHorizontalSigma(transfer_time);
	else if (wind_ > 3.)
	  sigma_y =
	    ComputeDouryNormalPlumeHorizontalSigma(transfer_time);
	else
	  sigma_y =
	    ComputeDouryLowPlumeHorizontalSigma(transfer_time);
      }
    else
      sigma_y = ComputeCrosswindSigma(transfer_time, z,
				      friction_velocity_,
				      convective_velocity_,
				      boundary_height_,
				      lmo_, coriolis_, option_hpdm,
				      stability_);
    // Vertical diffusion parameter.
    if (option_briggs)
      if (option_rural)
	if (!diff)
	  sigma_z =
	    ComputeRuralPlumeVerticalSigma(distance, stability_);
	else
	  sigma_z =
	    DifferentiateRuralPlumeVerticalSigma(distance, stability_);
      else
	if (!diff)
	  sigma_z =
	    ComputeUrbanPlumeVerticalSigma(distance, stability_);
	else
	  sigma_z =
	    DifferentiateUrbanPlumeVerticalSigma(distance, stability_);
    else if (option_doury)
      {
	if (option_day)
	  if (!diff)
	    sigma_z =
	      ComputeDouryNormalPlumeVerticalSigma(transfer_time);
	  else
	    sigma_z =
	      DifferentiateDouryNormalPlumeVerticalSigma(transfer_time);
	else if (wind_ > 3.)
	  if (!diff)
	    sigma_z =
	      ComputeDouryNormalPlumeVerticalSigma(transfer_time);
	  else
	    sigma_z =
	      DifferentiateDouryNormalPlumeVerticalSigma(transfer_time);
	else
	  if (!diff)
	    sigma_z =
	      ComputeDouryLowPlumeVerticalSigma(transfer_time);
	  else
	    sigma_z =
	      DifferentiateDouryLowPlumeVerticalSigma(transfer_time);
      }
    else
      if (!diff)
	sigma_z = ComputeVerticalSigma(transfer_time, z,
				       friction_velocity_,
				       convective_velocity_,
				       boundary_height_, lmo_, coriolis_,
				       option_hpdm, temperature_, stability_);
      else
	sigma_z = DifferentiateVerticalSigma(transfer_time, z, wind_,
					     friction_velocity_,
					     convective_velocity_,
					     boundary_height_,
					     lmo_, coriolis_, option_hpdm,
					     temperature_, stability_);
  }


  //!  Computes the loss factor.
  template<class T>
  void GaussianPlume<T>::ComputeLossFactor(T distance, T z, int species,
					   T& loss_factor, T& overcamp_factor)
  {
    T rad, bio, scav, dep;
    T transfer_time = distance / wind_;
    if (option_radioactive_decay)
      rad = half_life_time(species);
    else
      rad = 0.;
    if (option_biological_decay)
      bio = biological_half_life_time(species);
    else
      bio = 0.;
    if (option_scavenging)
      scav = scavenging_coefficient(species);
    else
      scav = 0.;
    if (option_dry_deposition)
      dep = deposition_velocity(species);
    else
      dep = 0.;

    ComputeLossFactor(distance, transfer_time, z, rad, bio, scav, dep,
		      loss_factor, overcamp_factor);
  }


  template<class T>
  void GaussianPlume<T>::ComputeLossFactor(T distance, T transfer_time,
					   T z, T rad, T bio, T scav, T dep,
					   T& loss_factor, T& overcamp_factor)
  {
    T radioactive_factor = 1.;
    T biological_factor = 1.;
    T scavenging_factor = 1.;
    T chamberlain_factor = 1.;
    overcamp_factor = 1.;

    const T pi = 3.14159265358979323846264;
    // log(2.)
    const T log2 = 0.69314718055994529;

    // Radioactive decay.
    if (option_radioactive_decay && rad != 0.)
      radioactive_factor = exp(-log2 * transfer_time / rad);

    // Biological decay.
    if (option_biological_decay && bio != 0.)
      biological_factor = exp(-log2 * transfer_time / bio);

    // Scavenging.
    if (option_scavenging)
      scavenging_factor = exp(- scav * transfer_time);

    // Dry deposition.
    if (option_dry_deposition)
      if (deposition_model == "Chamberlain")
	{
	  T delta_x = distance / T(Nchamberlain);
	  T z2 = 0.5 * z * z;
	  T dsigma_y, dsigma_z;
	  T x(0.);
	  T exp_factor = - dep / wind_ * sqrt(2. / pi);
	  
	  T fdry = 0;
	  for (int i = 1; i < Nchamberlain + 1; i++)
	    {
	      x += delta_x;
	      T time = x / wind_;
	      ComputeSigma(x, time, z, dsigma_y, dsigma_z);
	      fdry += exp(-z2 / (dsigma_z * dsigma_z)) / dsigma_z;
	    }
	  chamberlain_factor = exp(exp_factor * fdry * delta_x);
	}
      else
	{
	  T dsigma_z, sigma_y, sigma_z;
	  bool diffz = 1;
	  ComputeSigma(distance, transfer_time, z, sigma_y, sigma_z);
	  ComputeSigma(distance, transfer_time, z, sigma_y, dsigma_z, diffz);
	  overcamp_factor = 1. - 2. * dep /
	    (dep + wind_* z * dsigma_z / sigma_z);
	}
    loss_factor = radioactive_factor * scavenging_factor *
      biological_factor * chamberlain_factor;
  }


  //! Computes concentrations in the whole domain.
  template<class T>
  void GaussianPlume<T>::Compute()
  {
    this->InitCompute();

    int i, j, k, species;

    for (species = 0; species < this->Ns; species++)
      for (k = 0; k < this->Nz; k++)
        for (j = 0; j < this->Ny; j++)
          for (i = 0; i < this->Nx; i++)
            this->Concentration(species, k, j, i)
              = GetConcentration(species, this->GridZ4D(k),
                                 this->GridY4D(j), this->GridX4D(i));
  }


  //! Erases last sources of the list SourceList.
  template<class T>
  void GaussianPlume<T>::EraseSource()
  {
    for(int i = Nsource_temp; i != Nsource; i--)
      {
        typename list<PlumeSource<T>* >::iterator iter
          = SourceList.end();
        iter--;
        delete *iter;
        SourceList.erase(iter);
      }
    if(int(SourceList.size()) != Nsource)
      throw string("The number of sources must be ") + to_str<int>(Nsource)
        + " but is " + to_str<int>(SourceList.size()) + ".";
    Nsource_temp = Nsource;
  }


  //! Restore old rates.
  template<class T>
  void GaussianPlume<T>::RestoreRate()
  {
    T combination_factor, rate;
    for (typename list<PlumeSource<T>* >::const_iterator iter
	   = SourceList.begin();
	 iter != SourceList.end(); iter++)
      {
	if((*iter)->GetEmissionType() == "continuous_line")
	  {
	    combination_factor = (*iter)->GetCombinationFactor();
	    rate = (*iter)->GetRate();
	    if(combination_factor != 1)
	      (*iter)->SetRate(rate / (1 - combination_factor));
	  }
      }
  }


  //! Discretize a line source.
  template<class T>
  void GaussianPlume<T>::Discretization()
  {
    // Coefficient for the combination between point sources and line sources
    T combination_factor = 0;
    const T pi = 3.14159265358979323846264;
    T angle = wind_angle_;

    // Compute the value of wind_angle_ between 0 and 360 degrees.
    if (angle < 0)
      while(angle < 0)
        angle = angle + 360;
    else
      if (angle > 360)
        while(angle > 0)
          angle = angle - 360;

    // combination_factor varies linearly between 0 and 1 when the wind
    // becomes parallel to the road.
    // If combination_factor = 1 only point sources are used.
    // If combination_factor = 0 only line sources are used.

    // Angle between 80 and 90 degrees.
    if(sin_angle_ >= sin(80. * pi / 180.) && cos_angle_ >= 0)
      combination_factor = angle / 10 - 8.;
    // Angle between 90 and 100 degrees.
    else if(sin_angle_ >= sin(100. * pi / 180.) && cos_angle_ < 0)
      combination_factor = - angle / 10 + 10;
    // Angle between 260 and 270 degrees.
    else if(sin_angle_ <= sin(260. * pi / 180.) && cos_angle_ <= 0)
      combination_factor = angle / 10 - 26.;
    // Angle between 270 and 280 degrees.
    else if(sin_angle_ <= sin(280. * pi / 180.) && cos_angle_ > 0)
      combination_factor = - angle / 10 + 28.;
    else
      throw string("The line source should not be discretized.");

    Nsource_temp = Nsource;
    for (typename list<PlumeSource<T>* >::const_iterator iter
           = SourceList.begin();
         iter != SourceList.end(); iter++)
      {
        // If we have a line source then it is discretized.
        if((*iter)->GetEmissionType() == "continuous_line")
          {
            T x, y, z, x1, y1, z1, x2, y2, z2, rate, velocity,
              temperature, diameter, length, delta_x, delta_y, delta_z;
            int s, Np;

            x1 = (*iter)->GetX();
            y1 = (*iter)->GetY();
            z1 = (*iter)->GetZ();
            x2 = (*iter)->GetX2();
            y2 = (*iter)->GetY2();
            z2 = (*iter)->GetZ2();
            
            length = sqrt((x2-x1) * (x2-x1) + (y2-y1) * (y2-y1)
                          + (z2-z1) * (z2-z1));
            Np = int(length) + 1;

            delta_x = (x2-x1) / (Np - 1);
            delta_y = (y2-y1) / (Np - 1);
            delta_z = (z2-z1) / (Np - 1);

            for (int j = 0; j < Np; j++)
              {
                // Incrementation of coordinates.
                x = x1 + float(j) * delta_x;
                y = y1 + float(j) * delta_y;
                z = z1 + float(j) * delta_z;

                // Rate is multiplied by the combination_factor.
                rate = (*iter)->GetRate() * length * combination_factor / Np;
                velocity = (*iter)->GetVelocity();
                temperature = (*iter)->GetTemperature();
                diameter = (*iter)->GetDiameter();
                s = (*iter)->GetSpeciesIndex();
                PlumeSource<T>* ContinuousSource =
                  new PlumeSource<T>(rate, velocity, temperature,
                                     diameter, x, y, z, s);
                (ContinuousSource)->SetSigma_z(pow(0.7 * z,2));
                SourceList.push_back(ContinuousSource);
                Nsource_temp++;
              }
            /* The rate for the line source is multiplied
               by the combination_factor.*/
            (*iter)->SetRate((1 - combination_factor) * (*iter)->GetRate());
            (*iter)->SetCombinationFactor(combination_factor);
          }
      }
  }


  //!  Computes the correction factor for the downwind correction.
  template<class T>
  T GaussianPlume<T>::ComputeCorrection(T concentration, T cos_theta,
                                        T sin_theta, T tan_theta, T y1_source,
                                        T y2_source, T y_source, T  x_source,
                                        T distance_x, T effective_height,
                                        T sigma_y)
  {
    const T pi = 3.14159265358979323846264;
    T correction = 0;

    if(ReceptorDownwindPlume(x_source, y_source, y1_source,
                             y2_source, cos_theta, sin_theta))
      {
        // Scale to fit the estimation.
        distance_x = 10 * distance_x;

        // Wind angle in degree.
        T tan_theta = sin_theta / cos_theta;
        int id = 0, angle;

        if(cos_theta <= 0)
          angle = floor(atan(tan_theta) * 180 / pi + 0.5 + 180);
        else if (sin_theta >= 0)
          angle = floor(atan(tan_theta) * 180 / pi + 0.5);
        else
          angle = floor(atan(tan_theta) * 180 / pi + 0.5 + 360);

        if(abs(sin_theta) >=
           sin(threshold_downwind_correction_1_ * pi / 180.))
          {
            // If theta between threshold_downwind_correction_1_
            // and threshold_downwind_correction_2_.
            if (abs(sin_theta) <=
                sin(threshold_downwind_correction_2_ * pi / 180.))
              {
                // theta in the first quadrant.
                if(cos_theta > 0 && sin_theta > 0)
                  id = angle - threshold_downwind_correction_1_;
                // theta in the second quadrant.
                else if(cos_theta < 0 && sin_theta > 0)
                  id = 180 - threshold_downwind_correction_1_ - angle;
                // theta in the third quadrant.
                else if(cos_theta < 0 && sin_theta < 0)
                  id = angle - 180 - threshold_downwind_correction_1_;
                // theta in the fourth quadrant.
                else if(cos_theta > 0 && sin_theta < 0)
                  id = 360 - threshold_downwind_correction_1_ - angle;

                if (id > threshold_downwind_correction_2_ -
                    threshold_downwind_correction_1_ + 1)
                  throw "Index out of bounds for the correction";

                correction = (max_1_[id] * exp(-0.5 * pow((mu_1_[id] -
                                                           distance_x)
                                                          / sigma_1_[id],2)) +
                              max_2_[id] * exp(-0.5 * pow((mu_2_[id] -
                                                           distance_x)
                                                          /sigma_2_[id],2)));
              }
            // Else theta is between threshold_downwind_correction_2_
            // and 90 degrees.
            else
              {
                // theta in the first quadrant.
                if(cos_theta >= 0 && sin_theta > 0)
                  id = angle - threshold_downwind_correction_2_ - 1;
                // theta in the second quadrant.
                else if(cos_theta < 0 && sin_theta > 0)
                  id = 180 - threshold_downwind_correction_2_ - 1 - angle;
                // theta in the third quadrant.
                else if(cos_theta <= 0 && sin_theta < 0)
                  id = angle - 180 - threshold_downwind_correction_2_ - 1;
                // theta in the fourth quadrant.
                else if(cos_theta > 0 && sin_theta < 0)
                  id = 360 - threshold_downwind_correction_2_ - 1 - angle;

                if (id > 90 - threshold_downwind_correction_2_ + 1)
                  throw "Index out of bounds for the correction";

                if (distance_x < mu_[id])
                  correction =  (a_[id] + d_[id]) *
                    ( 1 - exp(- b_[id] * distance_x)) - d_[id];
                else
                  correction =  beta_[id]
                    * exp(-c_[id] * (distance_x - mu_[id]));
              }
          }
      }
    return correction;
  }


  //! Return true if the receptor is downwind of the source.
  template<class T>
  bool GaussianPlume<T>::ReceptorDownwindPlume(T x_source, T y_source,
					       T y1_source, T y2_source,
					       T cos_theta, T sin_theta)
  {
    bool res = 0;
    T tan_theta = sin_theta / cos_theta;//

    // Theta in the first quadrant.
    if (cos_theta > 0 && sin_theta >= 0)
      {
        if (tan_theta * x_source + y1_source < y_source &&
            tan_theta * x_source  + y2_source > y_source && x_source >= 0)
          res = 1;
      }
    // Theta in the second quadrant.
    else if(cos_theta < 0 && sin_theta > 0)
      {
        if (tan_theta * x_source + y1_source < y_source &&
            tan_theta * x_source  + y2_source > y_source && x_source <= 0)
          res = 1;
      }
    // Theta in the third quadrant.
    else if(cos_theta < 0 && sin_theta < 0)
      {
        if (tan_theta * x_source + y1_source < y_source &&
            tan_theta * x_source  + y2_source > y_source && x_source <= 0)
          res = 1;
      }
    // Theta in the fourth quadrant.
    else if(cos_theta > 0 && sin_theta < 0)
      {
        if (tan_theta * x_source + y1_source < y_source &&
            tan_theta * x_source  + y2_source > y_source && x_source >= 0)
          res = 1;
      }
    return res;
  }


  //!  Computes the extremities correction factors.
  template<class T>
  T GaussianPlume<T>::ComputeExtremitiesCorrection(T rate, T loss_factor,
                                                   T wind,T inversion_height_,
                                                   T effective_height,
                                                   T distance_x,
                                                   T distance_y,
                                                   T z, T initial_sigma_z_2,
                                                   T overcamp_factor,
                                                   T minimum_volume,
                                                   T cos_theta, int species)
  {
    T extremity = 0, sigma_y, sigma_z,
      time, initial_sigma_y, velocity = 10.1,
      section = 1.25664e-05, initial_sigma_y_2;
    const T pi = 3.14159265358979323846264;

    // Corrective coeficients based on estimations.
    T Lambda = 0;
    for (int i = 0;i < int(lambda_.size());i++)
      Lambda = Lambda + lambda_[i] * pow(cos_theta,i);
    Lambda = exp(Lambda);

    rate = rate / Lambda;

    initial_sigma_y_2 = section / pi;

    // Computing standard deviation sigma_y, with
    // distance_x.
    time = distance_x / wind;
    ComputeSigma(distance_x, time, effective_height,
                 sigma_y, sigma_z);
    sigma_y = sqrt(sigma_y * sigma_y +
		   initial_sigma_y_2);
    if (option_gillani && effective_height
        >= boundary_height_)
      {
        initial_sigma_z_2 = max(initial_sigma_z_2, 9.);
        sigma_z = sqrt(initial_sigma_z_2
                       * (1. + 2.3 * sqrt(time)));
      }
    else
      sigma_z = sqrt(sigma_z * sigma_z
                     + initial_sigma_z_2);

    if (distance_x > 0 && effective_height != 0.)
      {
        minimum_volume = section  * velocity * 4;
        
        // Computing loss factor.
        ComputeLossFactor(distance_x, effective_height,
                          species, loss_factor,
                          overcamp_factor);
        extremity = loss_factor
          * ComputePlumeConcentration(wind, inversion_height_,
                                      effective_height, rate, distance_y, z,
                                      sigma_y, sigma_z, overcamp_factor,
                                      minimum_volume);
        
      }
    return extremity;
  }


  //!  Read correction coefficients from the correction coefficients file.
  template<class T>
  vector<T> GaussianPlume<T>::ReadCoefficients(string section, int length)
  {
    vector<T> coef;
    vector<string> coefficients_line;
    ConfigStreams coefficients(correction_coefficient_path_);

    if (option_rural)
      coefficients.Find("[rural]");
    else
      coefficients.Find("[urban]");
    coefficients.Find("[" + stability_class_ + "]");
    coefficients.Find(section);
    coefficients_line = split(coefficients.GetLine());
    for (int i = 0;i < length;i++)
      coef.push_back(to_num<T>(coefficients_line[i]));
    return coef;
  }

  
  template<class T>
  void GaussianPlume<T>::InitCorrectionCoefficients()
  {
    int id_1, id_2;
    // Different range of angles for the correction regarding of
    // the stability class.
    threshold_downwind_correction_1_ = 0;
    threshold_downwind_correction_2_ = 0;

    if (option_rural)
      {
        if (stability_ == 0)
          {
            threshold_downwind_correction_1_ = 19;
            threshold_downwind_correction_2_ = 53;
            threshold_extremities_correction_ = 20;
          }
        else if (stability_ == 1)
          {
            threshold_downwind_correction_1_ = 23;
            threshold_downwind_correction_2_ = 60;
            threshold_extremities_correction_ = 30;
          }
        else if (stability_ == 2)
          {
            threshold_downwind_correction_1_ = 30;
            threshold_downwind_correction_2_ = 67;
            threshold_extremities_correction_ = 30;
          }
        else if (stability_ == 3)
          {
            threshold_downwind_correction_1_ = 40;
            threshold_downwind_correction_2_ = 74;
            threshold_extremities_correction_ = 45;
          }
        else if (stability_ == 4)
          {
            threshold_downwind_correction_1_ = 50;
            threshold_downwind_correction_2_ = 75;
            threshold_extremities_correction_ = 55;
          }
        else if (stability_ == 5)
          {
            threshold_downwind_correction_1_ = 71;
            threshold_downwind_correction_2_ = 81;
            threshold_extremities_correction_ = 70;
	  }
      }
    else
      {
        if (stability_ == 0 || stability_ == 1)
          {
            threshold_downwind_correction_1_ = 5;
            threshold_downwind_correction_2_ = 43;
            threshold_extremities_correction_ = 10;
          }
        else if (stability_ == 2)
          {
            threshold_downwind_correction_1_ = 9;
            threshold_downwind_correction_2_ = 54;
            threshold_extremities_correction_ = 10;
          }
        else if (stability_ == 3)
          {
            threshold_downwind_correction_1_ = 13;
            threshold_downwind_correction_2_ = 62;
            threshold_extremities_correction_ = 20;
          }
        else if (stability_ == 4 || stability_ == 5)
          {
            threshold_downwind_correction_1_ = 20;
            threshold_downwind_correction_2_ = 69;
            threshold_extremities_correction_ = 20;
          }
      }
    id_1 = threshold_downwind_correction_2_ -
      threshold_downwind_correction_1_ + 1;
    id_2 = 90 - threshold_downwind_correction_2_ - 1;

    // Correction coefficients for downwind receptors. Angles between
    // 40 degrees and 75 degrees.
    max_1_ = ReadCoefficients("[max_1]",id_1);
    mu_1_ = ReadCoefficients("[mu_1]",id_1);
    sigma_1_ = ReadCoefficients("[sigma_1]",id_1);
    max_2_ = ReadCoefficients("[max_2]",id_1);
    mu_2_ = ReadCoefficients("[mu_2]",id_1);
    sigma_2_ = ReadCoefficients("[sigma_2]",id_1);

    // Correction coefficients for downwind receptors. Angles between
    // 75 degrees and 90 degrees.
    mu_ = ReadCoefficients("[mu_]",id_2);
    a_ = ReadCoefficients("[a_]",id_2);
    b_ = ReadCoefficients("[b_]",id_2);
    d_ = ReadCoefficients("[d_]",id_2);
    beta_ = ReadCoefficients("[beta_]",id_2);
    c_ = ReadCoefficients("[b_]",id_2);

    // Coefficients for the correction at extremities of the source.
    lambda_ = ReadCoefficients("[lambda]",6);
  }
} // namespace Polyphemus.

#define POLYPHEMUS_FILE_MODELS_GAUSSIANPLUME_CXX
#endif
