// Copyright (C) 2006-2007, ENPC - INRIA - EDF R&D
// Author(s): Hadjira Foudhil, Vivien Mallet, Irène Korsakissok,
// Régis Briant
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is developed in the INRIA - ENPC joint project-team CLIME and in
// the ENPC - EDF R&D joint laboratory CEREA.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, visit the Polyphemus web site:
//      http://cerea.enpc.fr/polyphemus/

// This file is part of a Gaussian plume model for Polyphemus.


#ifndef POLYPHEMUS_FILE_MODELS_GAUSSIANPLUME_HXX


#include "PlumeSource.cxx"
#include <list>
#include "BaseModel.cxx"
#include "PlumeLineSource.cxx"


namespace Polyphemus
{


  using namespace std;


  ////////////////////
  // GAUSSIAN PLUME //
  ////////////////////


  //! Plume Gaussian model.
  template<class T>
  class GaussianPlume: public BaseModel<T>
  {

  protected:

    //! Read all input data, whatever the parameterizations may be?
    bool read_all;

    //! True if parameterization to compute standard deviations is "Briggs".
    bool option_briggs;
    //! True if parameterization to compute standard deviations is "Doury".
    bool option_doury;
    /*! \brief True if parameterization for sigma_z above BL is computed with
      Gillani. */
    bool option_gillani;
    //! True if HPDM formulae are used (only useful with similarity theory).
    bool option_hpdm;

    //! True if Holland formula for plume rise is used.
    bool option_holland;
    //! True if Concawe formula for plume rise is used.
    bool option_concawe;

    //! True if land category is set to "rural".
    bool option_rural;
    //! Is it daytime?.
    bool option_day;

    //! Is there radioactive decay?
    bool option_radioactive_decay;
    //! Is there biological decay?
    bool option_biological_decay;
    //! Is there scavenging?
    bool option_scavenging;
    //! Is there dry deposition?
    bool option_dry_deposition;

    //! Is there plume rise?
    bool option_plume_rise;
    //! Is there plume rise breakup?
    bool option_breakup;

    //! File storing all source data.
    string file_source;
    //! Source list.
    list<PlumeSource<T>* > SourceList;

    //! Number of sources.
    int Nsource;

    //! Temporary number of sources.
    int Nsource_temp;

    //! List of species half-lives (s).
    Array<T, 1> half_life_time;

    //! List of species biological half-lives (s).
    Array<T, 1> biological_half_life_time;

    //! List of species scavenging coefficients ( s^(-1) ).
    Array<T, 1> scavenging_coefficient;

    //! List of species deposition velocities (m/s).
    Array<T, 1> deposition_velocity;
    //! Deposition model ("Chamberlain" or "Overcamp").
    string deposition_model;
    //! Number of points to compute the Chamberlain integral.
    int Nchamberlain;

    //! Temperature of ambient air (Celsius degrees).
    T temperature_;
    //! Angle between positive x-axis and wind, counterclockwise (degrees).
    T wind_angle_;
    //! Wind (m/s).
    T wind_;
    //! Inversion height (m).
    T inversion_height_;
    //! Friction velocity (m/s).
    T friction_velocity_;
    //! Convective velocity (m/s).
    T convective_velocity_;
    //! Boundary layer height (m).
    T boundary_height_;
    //! Monin Obukhov length (m).
    T lmo_;
    //! Coriolis parameter.
    T coriolis_;

    //! Cosinus of wind angle.
    T cos_angle_;
    //! Sinus of wind angle.
    T sin_angle_;

    //! Stability class (A, B, C, D, E or F).
    string stability_class_;
    //! Stability (corresponding integer: from 0 for A to 5 for F).
    int stability_;

    // Correction coefficients for the line source model.
    vector<T> max_1_, mu_1_, sigma_1_,  max_2_, mu_2_, sigma_2_, mu_, a_, b_,
      c_, d_, beta_, lambda_;

    // Correction coefficients path file.
    string correction_coefficient_path_;

    // Threshold for the extremities correction in degrees.
    T threshold_extremities_correction_;

    // Threshold for the downsind correction in degrees.
    T threshold_downwind_correction_1_, threshold_downwind_correction_2_;

  public:

    /*** Main constructor ***/

    GaussianPlume(string config_file);

    /*** Access methods ***/

    bool GetBooleanOption(string option);
    void SetBooleanOption(string option, bool value);
    string GetStringOption(string option);
    void SetStringOption(string option, string value);
    T GetNumericalValue(string name);
    void SetNumericalValue(string name, T value);
    void GetSourceData(string name, Array<T, 1>& value);
    void SetSourceData(string name, const Array<T, 1>& value);
    void GetSpeciesData(string name, Array<T, 1>& value);
    void SetSpeciesData(string name, const Array<T, 1>& value);

    /*** Initializations ***/

    virtual void ReadConfiguration();
    virtual void Allocate();
    virtual void InitSource(string source_file);
    virtual void Init(bool read_all_input);
    virtual void InitMeteo(ConfigStream& meteo, bool show_meteo);
    virtual void InitScavenging(ConfigStream& meteo);
    virtual void InitDeposition(ConfigStream& meteo);

    /*** Computations ***/

    virtual void InitCompute();
    virtual void ComputePlumeRise();
    virtual void ComputeSourcePlumeRise(PlumeSource<T>& source);
    using BaseModel<T>::GetConcentration;
    virtual T GetConcentration(int species, T z, T y, T x);
    virtual void Compute();
    virtual void ComputeSigma(T distance, T transfer_time,
			      T z, T& sigma_y, T& sigma_z, bool diff = 0);
    void ComputeLossFactor(T distance, T z, int species,
			   T& loss_factor, T& overcamp_factor);
    void ComputeLossFactor(T distance, T transfer_time,
			   T z, T rad, T bio, T scav, T dep,
			   T& loss_factor, T& overcamp_factor);
    virtual void EraseSource();
    virtual void RestoreRate();
    virtual void Discretization();
    virtual T ComputeCorrection(T concentration, T cos_theta, T sin_theta,
                                T tan_theta, T y1_source, T y2_source,
                                T y_source, T  x_source, T distance_x,
                                T effective_height, T sigma_y);
    virtual bool ReceptorDownwindPlume(T x_source, T y_source, T y1_source,
				       T y2_source, T cos_theta, T sin_theta);
    virtual T ComputeExtremitiesCorrection(T rate, T loss_factor, T wind_,
                                           T inversion_height_,
                                           T effective_height,
                                           T distance_x, T distance_y,
                                           T z, T iniital_sigma_z_2,
                                           T overcamp_factor,T minimum_volume,
                                           T cos_theta, int species);
    virtual vector<T> ReadCoefficients(string section, int length);
    virtual void InitCorrectionCoefficients();
  };


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODELS_GAUSSIANPLUME_HXX
#endif
