C-----------------------------------------------------------------------
C     Copyright (C) 2007, ENPC - INRIA - EDF R&D
C     Author(s): Maryline Tombette
C     
C     This file is part of the Simple Aqueous model (SIMPLE_AQUEOUS), a
C     component of the air quality modeling system Polyphemus.
C    
C     Polyphemus is developed in the INRIA - ENPC joint project-team
C     CLIME and in the ENPC - EDF R&D joint laboratory CEREA.
C    
C     Polyphemus is free software; you can redistribute it and/or modify
C     it under the terms of the GNU General Public License as published
C     by the Free Software Foundation; either version 2 of the License,
C     or (at your option) any later version.
C     
C     Polyphemus is distributed in the hope that it will be useful, but
C     WITHOUT ANY WARRANTY; without even the implied warranty of
C     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
C     General Public License for more details.
C     
C     For more information, visit the Polyphemus web site:
C     http://cerea.enpc.fr/polyphemus/
C-----------------------------------------------------------------------

      subroutine electro_simple(x,conc_in,henrys,cst_dissoc,temp,lwc,
     $     f,frel,conc_out)
C------------------------------------------------------------------------
C     
C     -- DESCRIPTION 
C     
C     This routine computes the electroneutrality balance for the 
C     aqueous-phase model.
C------------------------------------------------------------------------
C     
C     -- INPUT VARIABLES
C     
C     X    : H+ concentration. [mol.L-1].
C     CONC_IN  : concentration vector:
C                conc_in(1): S(IV) ([mol/L_water])
C                conc_in(2): S(VI) ([mol/L_water])
C                conc_in(3): total ammonium ([atm])
C                conc_in(4): total nitrate ([atm])
C                conc_in(5): total CO2 ([atm])
C     cst_dissoc : equilibrium rate constants ([mol.L-1]).
C     henrys : Henry's rates ([mol.L-1.K-1]).
C     temp : temperature ([K])
C     lwc : liquid water content ([vol water / vol air])
C
C     -- INPUT/OUTPUT VARIABLES
C     
C     -- OUTPUT VARIABLES
C     
C     F : evaluation of the electroneutrality relation (anion-cation).
C     FL: relative value (ABS(F)/(anion+cation).
C     conc_out: output concentrations of ions ([mol/L_water]), 
C               partial pressures([atm]),
C               and aqueous species(mol/L_water]):
C               conc_out(1): [H+]
C               conc_out(2): [SO2(aq)]
C               conc_out(3): [HSO3-]
C               conc_out(4): [SO3^2-]
C               conc_out(5): [HNO3(aq)]
C               conc_out(6): [NO3-]
C               conc_out(7): [NH3(aq)]
C               conc_out(8): [NH4+]
C               conc_out(9): partial pressure HNO3(g)
C               conc_out(10): partial pressure NH3(g).
C     
C------------------------------------------------------------------------
C     
C     -- REMARKS
C     
C     
C------------------------------------------------------------------------
C     
C     -- MODIFICATIONS
C     
C------------------------------------------------------------------------
C     
C     -- AUTHOR(S)
C     
C     Marilyne Tombette, CEREA.
C     
C------------------------------------------------------------------------

      IMPLICIT NONE

      include 'droppar_simple.inc'
      include 'data_mass.inc'

      double precision conc_in(5),henrys(ngas_aq),conc_out(10)
      double precision x,temp,lwc
      double precision f,frel
      double precision cst_dissoc(nreact_dissoc)
      double precision cst_oxydation(nreact_oxydation)
      
      double precision cHSO3m, cSO32m, cHSO4m
      double precision cSO42m, cNO3m, cOHm, cHCO3m
      double precision cCO32m
      double precision cNH4p, ppso2, pphno3g,ppnh3g
      double precision fanion,fcation
      double precision so2aq, csiv,csvi,hno3aq,co2aq
      double precision nh3aq,h2so4aq

      double precision cst_RT

      cst_RT = 8.205d-2 * temp

C     1)Compute ions in aqueous-phase
C     
C     -----------------------------------

      cOHm=  cst_dissoc(1)/x                            !OH-

      pphno3g = conc_in(4) /
     $     (1 + cst_RT * lwc * henrys(ighno3)
     $     * (1 + cst_dissoc(7)/x)) ! in atm
      hno3aq = henrys(ighno3) * pphno3g ! in mol.L-1_water
      cNO3m = (cst_dissoc(7)/x) * hno3aq

      ppnh3g = conc_in(3) /
     $     (1 + cst_RT * lwc * henrys(ignh3)
     $     * (1 + cst_dissoc(6)*x/cst_dissoc(1))) ! in atm
      nh3aq = henrys(ignh3) * ppnh3g
      cNH4p = cst_dissoc(6)*x*nh3aq/cst_dissoc(1)

      SO2aq = conc_in(1) / ( 1 + cst_dissoc(2)/x +
     $     cst_dissoc(2)*cst_dissoc(3)/(x*x))
      cHSO3m = conc_in(1)/( 1 + x/cst_dissoc(2) 
     $     + cst_dissoc(3)/x) 
      cSO32m = conc_in(1)/ (1 + x/cst_dissoc(3) 
     $     + x*x / (cst_dissoc(2)*cst_dissoc(3)))

      h2so4aq = conc_in(2) / (1 + cst_dissoc(4)/x 
     $     + cst_dissoc(4)*cst_dissoc(5)/(x*x))

      cHSO4m= h2so4aq*cst_dissoc(4)/x !HSO4-
      cSO42m= cHSO4m*cst_dissoc(5)/x !SO4--

      co2aq = henrys(igco2) * conc_in(5)
     $     / (1+ lwc*cst_RT*henrys(igco2)
     $     * (1 + cst_dissoc(8)/x 
     $     + cst_dissoc(8)*cst_dissoc(9)/(x*x))) 
      cHCO3m = cst_dissoc(8)*co2aq/x 
      cCO32m = cst_dissoc(9)*cHCO3m/x

      conc_out(1)=x
      conc_out(2)=SO2aq
      conc_out(3)=cHSO3m
      conc_out(4)=cSO32m
      conc_out(5)=hno3aq
      conc_out(6)=cNO3m
      conc_out(7)=nh3aq
      conc_out(8)=cNH4p
      conc_out(9)=pphno3g
      conc_out(10)=ppnh3g

C     2) Compute electroneutrality balance
C     -----------------------------------

      fanion  = cHSO3m + 2.0d0*cSO32m + cHSO4m + 2.0d0*cSO42m
     $     + cNO3m + cOHm + cHCO3m + 2.0d0*cCO32m

      fcation = cNH4p + x
      f= fanion -fcation

      if (fanion+fcation.gt.0.d0) then
         frel=abs(f)/(fanion+fcation)
      else
         frel=-1.d0
      endif

      return
      end




