#include"libFlux.hxx"

template <class T> const T& min ( const T& a, const T& b ) {
  return !(b<a)?a:b;     // or: return !comp(b,a)?a:b; for the comp version
}

/*template <typename T> const T min(const T a, const T b)
{
  return (a<b)?a:b;
}
*/
/*****************************************************/
/*           Compute the deposition flux             */
/*               (roughness approach)                */
/*                                                   */
/* param:  sedimentation_velocity                    */
/*                        - sedimentation velocity   */
/*         Ra             - aerodynamical resistance */
/*         Rs             - surface resistance       */
/*                                                   */
/* return: RoughnessFlux       - flux in roughness   */
/*                                   approach        */
/*****************************************************/


float ComputeRoughnessFlux(float sedimentation_velocity, float Ra, float Rs)
{
  float Flux = sedimentation_velocity /
    ( 1. - exp(- sedimentation_velocity * ( Ra + Rs ) ) );

  return Flux;
}

/*****************************************************/
/*  Compute the canyon's concentration (over air     */
/*   concentration) for any canyon's geometry        */
/*                                                   */
/* param:  l_can          - canyon's width           */
/*         l_wall         - walls' height            */
/*         l_street       - road's width             */
/*         sedimentation_velocity                    */
/*                        - sedimentation velocity   */
/*         Ra_canyon      - aerodynamical resistance */
/*                           of the canyon           */
/*         Ra_street      - aerodynamical resistance */
/*                           of the street           */
/*         Rs_street      - surface resistance       */
/*                           of the street           */
/*         Ra_wall        - aerodynamical resistance */
/*                           of the wall             */
/*         Rs_wall        - surface resistance       */
/*                           of the wall             */
/*                                                   */
/* return: Ccan           - canyon's concentration   */
/*****************************************************/

float ComputeCcan(float l_can, float l_wall, float l_street, 
		  float sedimentation_velocity, float Ra_canyon,
		  float Ra_street, float Rs_street, float Ra_wall,
		  float Rs_wall)
{
  /*
  float coeff = 1. - exp(- sedimentation_velocity * Ra_canyon);

  coeff = coeff *
    ( ( l_wall / l_can ) / ( sedimentation_velocity * ( Ra_wall + Rs_wall ) )
    + ( l_street / l_can ) / ( 1. - exp(- sedimentation_velocity * ( Ra_street + Rs_street ) ) )
    + 1. / ( exp(sedimentation_velocity * Ra_canyon) - 1. )
      );
  */

  float coeff = ( l_wall / l_can )
    * ( 1. - exp(- sedimentation_velocity * Ra_canyon)  )
    / ( sedimentation_velocity * ( Ra_wall + Rs_wall ) )
    + ( l_street / l_can )
    * ( 1. - exp(- sedimentation_velocity * Ra_canyon)  )
    / ( 1. - exp(- sedimentation_velocity * ( Ra_street + Rs_street ) ) )
    + exp(- sedimentation_velocity * Ra_canyon);

  coeff = 1. / coeff;

  return coeff;

}

/*****************************************************/
/*    Compute the deposition flux over the street    */
/*                                                   */
/* param:  l_can          - canyon's width           */
/*         l_wall         - walls' height            */
/*         l_street       - road's width             */
/*         sedimentation_velocity                    */
/*                        - sedimentation velocity   */
/*         Ra_canyon      - aerodynamical resistance */
/*                           of the canyon           */
/*         Ra_street      - aerodynamical resistance */
/*                           of the street           */
/*         Rs_street      - surface resistance       */
/*                           of the street           */
/*         Ra_wall        - aerodynamical resistance */
/*                           of the wall             */
/*         Rs_wall        - surface resistance       */
/*                           of the wall             */
/*                                                   */
/* return: StreetFlux     - street flux              */
/*****************************************************/

float ComputeStreetFlux(float l_can, float l_wall, float l_street, 
			float sedimentation_velocity, float Ra_canyon,
			float Ra_street, float Rs_street, float Ra_wall,
			float Rs_wall)
{
  float Ccan = ComputeCcan(l_can, l_wall, l_street, 
			   sedimentation_velocity, Ra_canyon,
			   Ra_street, Rs_street, Ra_wall, Rs_wall);

  float StreetFlux = Ccan * sedimentation_velocity /
    ( 1. - exp(- sedimentation_velocity * (Ra_street + Rs_street) ) );

  return StreetFlux;
}

/*****************************************************/
/*    Compute the deposition flux over the wall      */
/*                                                   */
/* param:  l_can          - canyon's width           */
/*         l_wall         - walls' height            */
/*         l_street       - road's width             */
/*         sedimentation_velocity                    */
/*                        - sedimentation velocity   */
/*         Ra_canyon      - aerodynamical resistance */
/*                           of the canyon           */
/*         Ra_street      - aerodynamical resistance */
/*                           of the street           */
/*         Rs_street      - surface resistance       */
/*                           of the street           */
/*         Ra_wall        - aerodynamical resistance */
/*                           of the wall             */
/*         Rs_wall        - surface resistance       */
/*                           of the wall             */
/*                                                   */
/* return: WallFlux       - wall flux                */
/*****************************************************/

float ComputeWallFlux(float l_can, float l_wall, float l_street, 
		      float sedimentation_velocity, float Ra_canyon,
		      float Ra_street, float Rs_street, float Ra_wall,
		      float Rs_wall)
{
  float Ccan = ComputeCcan(l_can, l_wall, l_street, 
			   sedimentation_velocity, Ra_canyon,
			   Ra_street, Rs_street, Ra_wall, Rs_wall);

  float WallFlux = Ccan / ( Ra_wall + Rs_wall ) ;

  return WallFlux;
}

/*****************************************************/
/*    Compute the deposition flux over the roof      */
/*                                                   */
/* param:  sedimentation_velocity                    */
/*                        - sedimentation velocity   */
/*         Ra_roof        - aerodynamical resistance */
/*                           of the roof             */
/*         Rs_roof        - surface resistance       */
/*                           of the roof             */
/*                                                   */
/* return: RoofFlux       - roof flux                */
/*****************************************************/

float ComputeRoofFlux(float sedimentation_velocity, float Ra_roof,
		      float Rs_roof)
{
  float RoofFlux = sedimentation_velocity /
    ( 1. - exp(- sedimentation_velocity * (Ra_roof + Rs_roof) ) );

  return RoofFlux;
}


/*****************************************************/
/*    Compute the equivalent deposition flux         */
/*         removed from the atmosphere               */
/*                                                   */
/* param:  lambda_p - Plan area density              */
/*         w        - canyon's mean width            */
/*         h        - buildings' mean height         */
/*         StreetFlux_free                           */
/*                  - flux over the street in the    */
/*                      free region                  */
/*         WallFlux_free                             */
/*                  - flux over the wall(s) in the   */
/*                      free region                  */
/*         StreetFlux_recirc                         */
/*                  - flux over the street in the    */
/*                      recirculation region         */
/*         WallFlux_recirc                           */
/*                  - flux over the wall(s) in the   */
/*                      recirculation region         */
/*         RoofFlux - flux over roof                 */
/*                                                   */
/* return: FluxEquiv       - deposition flux removed */
/*                            from the atmosphere    */
/*****************************************************/

float ComputeFluxEquiv(float lambda_p, float h, float w,
		       float StreetFlux_free, float WallFlux_free,
		       float StreetFlux_recirculation,
		       float WallFlux_recirculation, float RoofFlux)
{
  float FluxEquiv_free;
  float FluxEquiv_recirculation;

  float w_r = 3. * h;
  float w_r_2 = 3. * h / 2.;

  float l_wall_free;
  float l_wall_recirculation;

  float l_can_free = w - min( w_r_2, w );
  //float l_can_recirculation = min( w_r_2, w );
  float l_street_free = w - min( w_r, w );
  float l_street_recirculation = min( w_r, w );
  if(w < w_r_2)
    {
      l_wall_free = 0.;
      l_wall_recirculation = 2. * h;
    }
  else if(w > w_r_2 && w < w_r)
    {
      //l_wall_free = h * w / w_r;
      l_wall_free = 2. * h * ( w / w_r - 0.5 );
      //l_wall_recirculation = h * ( 2. - w / w_r );
      l_wall_recirculation = 2. * h * ( 1.5 - w / w_r );
    }
  // w > w_r
  else
    {
      l_wall_free = h;
      l_wall_recirculation = h;
    }

  // If free region exists (w > w_r /2.)
  if(l_can_free > 0)
    FluxEquiv_free = ComputeRegionFluxEquiv(lambda_p, w, l_wall_free, l_street_free,
					    StreetFlux_free, WallFlux_free);
    /*
    FluxEquiv_free = ComputeFreeFlux(lambda_p, w, h,
				     l_can_free, l_wall_free,
				     l_street_free,
				     StreetFlux_free,
				     WallFlux_free);

    */
  else
    FluxEquiv_free = 0;


  //if(l_can_recirculation >0)
  /*
  FluxEquiv_recirculation = ComputeRecirculationFlux(lambda_p, w, h,
						       l_can_recirculation,
						       l_wall_recirculation,
						       l_street_recirculation,
						       StreetFlux_recirculation,
						       WallFlux_recirculation);
  */
  //else
  //FluxEquiv_recirculation = 0;

  FluxEquiv_recirculation = ComputeRegionFluxEquiv(lambda_p, w, l_wall_recirculation,
						   l_street_recirculation,
						   StreetFlux_recirculation,
						   WallFlux_recirculation);

  float FluxEquiv = lambda_p * RoofFlux
    + FluxEquiv_free + FluxEquiv_recirculation;

  return FluxEquiv;
}

float ComputeFluxEquiv(float lambda_p, float h, float w, float w_r,
		       float StreetFlux_free, float WallFlux_free,
		       float StreetFlux_recirculation,
		       float WallFlux_recirculation, float RoofFlux)
{
  float FluxEquiv_free;
  float FluxEquiv_recirculation;

  float w_r_2 = w_r / 2.;

  float l_wall_free;
  float l_wall_recirculation;

  float l_can_free = w - min( w_r_2, w );
  //float l_can_recirculation = min( w_r_2, w );
  float l_street_free = w - min( w_r, w );
  float l_street_recirculation = min( w_r, w );
  if(w < w_r_2)
    {
      l_wall_free = 0.;
      l_wall_recirculation = 2. * h;
    }
  else if(w > w_r_2 && w < w_r)
    {
      //l_wall_free = h * w / w_r;
      l_wall_free = 2. * h * ( w / w_r - 0.5 );
      //l_wall_recirculation = h * ( 2. - w / w_r );
      l_wall_recirculation = 2. * h * ( 1.5 - w / w_r );
    }
  // w > w_r
  else
    {
      l_wall_free = h;
      l_wall_recirculation = h;
    }

  // If free region exists (w > w_r /2.)
  if(l_can_free > 0)
    FluxEquiv_free = ComputeRegionFluxEquiv(lambda_p, w, l_wall_free, l_street_free,
					    StreetFlux_free, WallFlux_free);
  else
    FluxEquiv_free = 0;

  FluxEquiv_recirculation = ComputeRegionFluxEquiv(lambda_p, w, l_wall_recirculation,
						   l_street_recirculation,
						   StreetFlux_recirculation,
						   WallFlux_recirculation);

  float FluxEquiv = lambda_p * RoofFlux
    + FluxEquiv_free + FluxEquiv_recirculation;

  return FluxEquiv;
}


/*****************************************************/
/*   Compute the equivalent deposition flux removed  */
/*     from the atmosphere to a particular region    */
/*                                                   */
/* param:  lambda_p - plan area density              */
/*         h        - buildings' mean height         */
/*         w        - canyon's mean width            */
/*         l_can    - canyon's width (region)        */
/*         l_wall   - walls' height (region)         */
/*         l_street - road's width (region)          */
/*         StreetFlux                                */
/*                  - flux over the street (region)  */
/*         WallFlux                                  */
/*                  - flux over the wall(s) (region) */
/*                                                   */
/* return: RegionFluxEquiv - deposition flux removed */
/*                              from the atmosphere  */
/*****************************************************/

float ComputeRegionFluxEquiv(float lambda_p, float w, float l_wall, float l_street,
			     float StreetFlux, float WallFlux)
{
  float RegionFluxEquiv = ( 1. - lambda_p ) * ( l_street * StreetFlux + l_wall * WallFlux ) / w;

  return RegionFluxEquiv;
}
